#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Fast5 compression filter for demonstrating vbz compression
"""

import os
import h5py
import argparse
from shutil import copy2

__version__ = "0.1.1"


def compress_fast5(filename, output_suffix, vbz_version=0, decompress=False):
    """
    (De)compress the raw signal in the fast5
    """
    # create a copy of the input file for compressing
    filename = os.path.abspath(filename)
    out_filename = filename + output_suffix
    copy2(filename, out_filename)

    # open the copy and compress the signal datasets only
    with h5py.File(out_filename, 'r+') as fast5:

        if decompress:
            compression_opts = {
                "compression": "gzip", "compression_opts": 1
            }
        else:
            compression_opts = {
                "compression": 32020,
                # 2 byte integers with zig zag + level 1 zstd
                "compression_opts": (vbz_version, 2, 1, 1) 
            }

        # multi read files
        for read in fast5:
            if not read.startswith("read_"):
                continue
            
            # read the signal dataset
            group = '%s/Raw/' % read
            raw = fast5[group]["Signal"][:]
            
            # remove the dataset
            del fast5[group]["Signal"]

            # store the dataset with the new compression
            fast5[group].create_dataset(
                "Signal",
                data=raw,
                **compression_opts,
                chunks=(len(raw),)
            )

    return out_filename


def main(args, print_help):
    for filename in args.files:
        print(compress_fast5(filename, args.output_suffix, args.vbz_version, args.decompress))


if __name__ == "__main__":
    parser = argparse.ArgumentParser("fast5compress")
    parser.add_argument('files', nargs='*', help='input files')
    parser.add_argument("-d", "--decompress", action="store_true", default=False)
    parser.add_argument("-s", "--output-suffix", default=".tmp")
    parser.add_argument('-v', '--version', action='version', version=__version__)
    parser.add_argument('--vbz-version', type=int, default=1)
    main(parser.parse_args(), parser.print_help)



from setuptools import setup, find_packages

setup(
    name='pyvbz',
    version='1.0.1',
    url='https://nanoporetech.com',
    author='Oxford Nanopore Technologies, Limited',
    author_email="support@nanoporetech.com",
    packages=find_packages(),
    description="Python bindings to libvbz",
    install_requires=['numpy', 'cffi'],
    setup_requires=['cffi', 'wheel'],
    cffi_modules=['vbz/build.py:ffibuilder'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Mozilla Public License 2.0',
        'Natural Language :: English',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
    ]
)



import h5py
import matplotlib.pyplot as plt
import numpy
import time

import zstd


def time_dataset(data, create_fn, target_time=0.1):
    element_size = data.dtype.itemsize
    total_time = 0
    times = []
    count = 0

    for _ in range(1000):
        
        begin = time.time()
        with h5py.File("test_file.h5", "w") as f:
            create_fn(f, "bar", data, element_size)
        end = time.time()

        count += 1
        this_time = (end - begin)
        times.append(this_time)
        total_time += this_time

        if total_time > target_time:
            break

    return sorted(times)[len(times)//2]

def create_uncompressed(f, name, data, element_size):
    f.create_dataset(name, data=data, chunks=(len(data),),)

def create_vbz_zstd(f, name, data, element_size):
    use_zig_zag = 1
    f.create_dataset("bar", compression=32020, compression_opts=(0, element_size, use_zig_zag, 1), data=data, chunks=(len(data),))

def create_vbz(f, name, data, element_size):
    use_zig_zag = 1
    f.create_dataset("bar", compression=32020, compression_opts=(0, element_size, use_zig_zag, 0), data=data, chunks=(len(data),))

def create_zlib(f, name, data, element_size):
    f.create_dataset("bar", compression="gzip", chunks=(len(data),), data=data)

def create_zstd_raw(f, name, data, element_size):
    zstd_data = numpy.frombuffer(zstd.compress(data), dtype="u1")
    f.create_dataset("bar", chunks=(len(zstd_data),), data=zstd_data)


if __name__ == "__main__":
    plots = []
    plot_names = []

    palette = plt.get_cmap('Set1')

    for m_i, method in enumerate([create_zstd_raw, create_vbz, create_vbz_zstd, create_zlib, create_uncompressed]):
        data_types = ["i1", "i2", "i4"]
        for d_i, data_type in enumerate(data_types):
            size_to_times = []

            for i in range(1,20):
                size = 1000000 * i
                data = numpy.random.randint(low=-50, high=50, size=size//int(data_type[1:]), dtype=data_type)
                size_to_times.append((data.nbytes, time_dataset(data, method)))

            print(size_to_times)
            x = [e[0] for e in size_to_times]
            y = [e[0]/e[1] for e in size_to_times]

            c_index = m_i
            
            plt.subplot(1, len(data_types), 1 + d_i)
            plots.append(plt.plot(x, y, color=palette(c_index))[0])
            plot_names.append("%s %s" % (method.__name__, data_type))

            #ax = plt.axes()
            #ax.grid()
            #ax.yaxis.set_major_formatter(plt.FuncFormatter(lambda v,_: format_size(v, "/s")))
            #ax.xaxis.set_major_formatter(plt.FuncFormatter(lambda v,_: format_size(v)))

            plt.xlabel("block size (bytes)")
            if d_i == 0:
                plt.ylabel("Write speed (bytes/second)")
            plt.legend(plots, plot_names)

        
    def format_size(val, suffix=""):
        units = ['bytes', 'KB', 'MB', 'GB']
        divisor = 1000

        unit_idx = 0
        while val > divisor and unit_idx < len(units):
            val /= divisor
            unit_idx += 1

        return '%.1f %s%s' % (val, units[unit_idx], suffix)

    plt.show()


import subprocess
import os
import sys

import unittest

TEST_DATA = os.path.join(os.path.dirname(__file__), "..", "..", "test_data")


class TestHDFRepack(unittest.TestCase):

    def test_simple_range(self):
        source_file = os.path.join(TEST_DATA, "multi_fast5_zip.fast5")
        dest_file_vbz = os.path.join(TEST_DATA, "multi_fast5_zip.fast5.repacked_vbz")
        dest_file_none = os.path.join(TEST_DATA, "multi_fast5_zip.fast5.repacked_none")

        result = subprocess.run([ 'h5repack', '-f', 'NONE', source_file, dest_file_none ],
            stdout=sys.stdout,
            stderr=sys.stderr
        )
        self.assertEqual(result.returncode, 0)

        result = subprocess.run([ 'h5repack', '-f', 'UD=32020,5,0,0,2,1,1', source_file, dest_file_vbz ],
            stdout=sys.stdout,
            stderr=sys.stderr
        )
        self.assertEqual(result.returncode, 0)

        uncompressed_bytes = os.stat(dest_file_none).st_size
        source_bytes = os.stat(source_file).st_size
        vbz_bytes = os.stat(dest_file_vbz).st_size
        print("Uncompressed file is %s bytes" 
            % uncompressed_bytes)
        print("Source file is %s bytes (%.0f%% of uncompressed)" 
            % (source_bytes, (source_bytes/uncompressed_bytes*100)))
        print("vbz file is %s bytes (%.0f%% of uncompressed, %.0f%% of source)" 
            % (vbz_bytes, (vbz_bytes/uncompressed_bytes*100), (vbz_bytes/source_bytes*100)))

if __name__ == '__main__':
    unittest.main()


"""
Builds the C Python wrapper for libvbz
"""

import os
from cffi import FFI


vbz_include_paths = os.environ["VBZ_INCLUDE_PATHS"].split(";")
vbz_libs = os.environ["VBZ_LINK_LIBS"].split(";")


ffibuilder = FFI()


ffibuilder.set_source(
    "_vbz",
    """
    #include <vbz.h>
    """,
    include_dirs=vbz_include_paths,
    extra_objects=vbz_libs,
    source_extension='.cpp',
    libraries=['c', 'stdc++'],
    extra_compile_args=['-std=c++11'],
)


ffibuilder.cdef("""
typedef uint32_t vbz_size_t;

typedef struct {
    bool perform_delta_zig_zag;
    unsigned int integer_size;
    unsigned int zstd_compression_level;
    unsigned int vbz_version;
} CompressionOptions;

bool vbz_is_error(vbz_size_t result_value);

vbz_size_t vbz_max_compressed_size(
    vbz_size_t source_size,
    CompressionOptions const* options
);

vbz_size_t vbz_decompressed_size(
    void const* source,
    vbz_size_t source_size,
    CompressionOptions const* options
);

vbz_size_t vbz_compress_sized(
    void const* source,
    vbz_size_t source_size,
    void* destination,
    vbz_size_t destination_capacity,
    CompressionOptions const* options
);

vbz_size_t vbz_decompress_sized(
    void const* source,
    vbz_size_t source_size,
    void* destination,
    vbz_size_t destination_capacity,
    CompressionOptions const* options
);
""")


if __name__ == "__main__":
    ffibuilder.compile()



"""
Python bindings for libvbz
"""

__version__ = "0.9.3"


import numpy as np

from _vbz import lib, ffi


def compression_options(zigzag, size, zlevel=1, version=0):
    options = ffi.new("CompressionOptions *")
    options.integer_size = size
    options.perform_delta_zig_zag = zigzag
    options.zstd_compression_level = zlevel
    options.vbz_version = version
    return options


def compress(data, options=None):

    if options is None:
        zigzag = np.issubdtype(data.dtype, np.signedinteger)
        options = compression_options(zigzag, data.dtype.itemsize)

    output_size = lib.vbz_max_compressed_size(len(data) * options.integer_size, options)
    output = np.empty(output_size, dtype=np.uint8)

    if lib.vbz_is_error(output_size):
        raise Exception("Something unexpected went wrong")

    size = lib.vbz_compress_sized(
        ffi.cast("void const *", ffi.from_buffer(data)),
        len(data) * options.integer_size,
        ffi.cast("void *", ffi.from_buffer(output)),
        output_size,
        options
    )

    if lib.vbz_is_error(size):
        raise Exception("Something unexpected went wrong")

    return output[:size]


def decompress(data, dtype, options=None):

    if options is None:
        zigzag = np.issubdtype(dtype, np.signedinteger)
        try:
            options = compression_options(zigzag, dtype.itemsize)
        except TypeError:
            options = compression_options(zigzag, dtype().itemsize)

    uncompressed_size = lib.vbz_decompressed_size(
        ffi.cast("void const *", ffi.from_buffer(data)),
        len(data),
        options
    )

    if lib.vbz_is_error(uncompressed_size):
        raise Exception("Something unexpected went wrong")

    output = np.empty(uncompressed_size, dtype=dtype)

    size = lib.vbz_decompress_sized(
        ffi.cast("void const *", ffi.from_buffer(data)),
        len(data),
        ffi.cast("void *", ffi.from_buffer(output)),
        uncompressed_size,
        options
    )

    if lib.vbz_is_error(size):
        raise Exception("Something unexpected went wrong")

    return output[:int(size / options.integer_size)]



import h5py
import numpy
import os

import unittest

TEST_DATA = os.path.join(os.path.dirname(__file__), "..", "..", "test_data")

class TestVbzFilter(unittest.TestCase):
    VBZ_VERISON = 0

    def test_simple_range(self):
        original_data = numpy.arange(1, 100, dtype="i2")

        with  h5py.File("test_file.h5", "w") as f:
            f.create_dataset("bar", compression=32020, compression_opts=(self.VBZ_VERISON, 2, 1, 1), data=original_data)

        with h5py.File("test_file.h5", "r") as f:
            numpy.testing.assert_array_equal(f["bar"][:], original_data)

    def test_compressing_real_data(self):
        zip_file = h5py.File(os.path.join(TEST_DATA, "multi_fast5_zip.fast5"))

        for zip_key in zip_file.keys():
            original_data = zip_file[zip_key]["Raw/Signal"][:]

            if (zip_key == "read_008468c3-e477-46c4-a6e2-7d021a4ebf0b"):
                continue

            with h5py.File("test_file.h5", "w") as f:
                f.create_dataset("bar", compression=32020, compression_opts=(self.VBZ_VERISON, 2, 1, 1), data=original_data, chunks=(len(original_data),))

            with h5py.File("test_file.h5", "r") as f:
                numpy.testing.assert_array_equal(f["bar"][:], original_data)

class TestVbzFilterV0(TestVbzFilter):
    VBZ_VERISON = 0
    
class TestVbzFilterV1(TestVbzFilter):
    VBZ_VERISON = 1
    
class TestVbzStoredFiles(unittest.TestCase):
    def test_stored_files(self):
        files = [
            h5py.File(os.path.join(TEST_DATA, "multi_fast5_zip.fast5")),
            h5py.File(os.path.join(TEST_DATA, "multi_fast5_vbz.fast5")),
            h5py.File(os.path.join(TEST_DATA, "multi_fast5_vbz_v1.fast5")),
        ]

        for sorted_keys in zip(*[sorted(file.keys()) for file in files ]):
            for e in sorted_keys[1:]:
                self.assertEqual(sorted_keys[0], e)

            key = sorted_keys[0]
            first_data = files[0][key]["Raw/Signal"][:]
            for file in files[1:]:
                data = file[key]["Raw/Signal"][:]
                numpy.testing.assert_array_equal(first_data, data)


if __name__ == '__main__':
    unittest.main()